// ************************************************************************** //
//
//  BornAgain: simulate and fit scattering at grazing incidence
//
//! @file      coregui/Models/ParameterizedItem.cpp
//! @brief     Implements class ParameterizedItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2015
//! @authors   Scientific Computing Group at MLZ Garching
//! @authors   C. Durniak, M. Ganeva, G. Pospelov, W. Van Herck, J. Wuttke
//
// ************************************************************************** //

#include "ParameterizedItem.h"
#include "Exceptions.h"
#include "MaterialEditor.h"
#include "GUIHelpers.h"
#include "PropertyVariantManager.h"
#include "GroupPropertyRegistry.h"
#include "GroupProperty.h"

#include <QEvent>
#include <QDynamicPropertyChangeEvent>
#include <QDebug>
#include <QMetaEnum>

const QString ParameterizedItem::P_NAME = "Name";
const QString ParameterizedItem::P_PORT = "Port";

ParameterizedItem::ParameterizedItem(QString model_type, ParameterizedItem *parent)
    : m_model_type(std::move(model_type)), mp_parent(parent)
{
    if (mp_parent) {
        mp_parent->insertChildItem(-1, this);
    }

    registerProperty(P_NAME, QString(), PropertyAttribute(PropertyAttribute::HIDDEN));
    registerProperty(P_PORT, -1, PropertyAttribute(PropertyAttribute::HIDDEN));
    setItemName(m_model_type);
}

ParameterizedItem::~ParameterizedItem()
{
    qDeleteAll(m_children);
    qDeleteAll(m_sub_items);
}

QString ParameterizedItem::modelType() const
{
    return m_model_type;
}

QString ParameterizedItem::itemName() const
{
    return getRegisteredProperty(P_NAME).toString();
}

void ParameterizedItem::setItemName(const QString &item_name)
{
    setRegisteredProperty(P_NAME, item_name);
}

QString ParameterizedItem::displayName() const
{
    return m_display_name;
}

void ParameterizedItem::setDisplayName(QString display_name)
{
    m_display_name = std::move(display_name);
}

QString ParameterizedItem::getItemLabel() const
{
    return QString("");
}

ParameterizedItem *ParameterizedItem::parent() const
{
    return mp_parent;
}

ParameterizedItem *ParameterizedItem::childAt(int row) const
{
    return m_children.value(row);
}

int ParameterizedItem::rowOfChild(ParameterizedItem *child) const
{
    return m_children.indexOf(child);
}

int ParameterizedItem::childItemCount() const
{
    return m_children.count();
}

bool ParameterizedItem::hasChildItems() const
{
    return !m_children.isEmpty();
}

QList<ParameterizedItem *> ParameterizedItem::childItems() const
{
    return m_children;
}

void ParameterizedItem::insertChildItem(int row, ParameterizedItem *item)
{
    if (row == -1)
        row = m_children.size();
    item->mp_parent = this;
    m_children.insert(row, item);
    onChildPropertyChange();
}

ParameterizedItem *ParameterizedItem::takeChildItem(int row)
{
    ParameterizedItem *item = m_children.takeAt(row);
    item->mp_parent = 0;
    onChildPropertyChange();
    return item;
}

ParameterizedItem *ParameterizedItem::getChildOfType(QString type) const
{
    for (QList<ParameterizedItem *>::const_iterator it = m_children.begin();
         it != m_children.end(); ++it) {
        if ((*it)->modelType() == type) return *it;
    }
    return 0;
}

bool ParameterizedItem::acceptsAsChild(const QString &child_name) const
{
    return m_valid_children.contains(child_name);
}

QList<QString> ParameterizedItem::acceptableChildItems() const
{
    return m_valid_children;
}

// emits signal on property change
bool ParameterizedItem::event(QEvent *e)
{
    if (e->type() == QEvent::DynamicPropertyChange) {
        QDynamicPropertyChangeEvent *propertyEvent = dynamic_cast<QDynamicPropertyChangeEvent *>(e);
        Q_ASSERT(e);
        QByteArray byte_array = propertyEvent->propertyName();
        QString name(byte_array.constData());
        onPropertyChange(name);
    }
    return QObject::event(e);
}

QMap<QString, ParameterizedItem *> ParameterizedItem::getSubItems() const
{
    return m_sub_items;
}

void ParameterizedItem::addPropertyItem(QString name, ParameterizedItem *item)
{
    // if (!item) return;
    Q_ASSERT(item);
    qDebug() << "ParameterizedItem::addPropertyItem()" << name;

    if (m_sub_items.contains(name)) {
        qDebug() << "       ParameterizedItem::addPropertyItem() -> item is already there" << name
                 << "replacing with " << item->modelType();
        delete m_sub_items[name];
        m_sub_items.remove(name);
    }
    m_sub_items[name] = item;
    item->mp_parent = this;
    onSubItemChanged(name);
    onChildPropertyChange();
    qDebug() << "ParameterizedItem::addPropertyItem() -> about to leave" << name;
}

bool ParameterizedItem::isRegisteredProperty(const QString &name)
{
    return m_registered_properties.contains(name);
}

ParameterizedItem *ParameterizedItem::registerGroupProperty(const QString &group_name,
                                                            const Constants::ModelType &group_model)
{
    qDebug() << "ParameterizedItem::registerGroupProperty() ->"
             << "this->modelType" << modelType() << "group_name" << group_name << " group_model"
             << group_model;

    GroupProperty_t group_property
        = GroupPropertyRegistry::createGroupProperty(group_name, group_model);
    QVariant variant;
    variant.setValue(group_property);
    registerProperty(group_name, variant);
    group_property->setParent(this);
    return m_sub_items[group_name];
}

ParameterizedItem *ParameterizedItem::setGroupProperty(const QString &name, const QString &value)
{
    qDebug() << "ParameterizedItem::setGroupProperty()" << name << value;
    GroupProperty_t group_property = getRegisteredProperty(name).value<GroupProperty_t>();
    group_property->setCurrentType(value);
    return m_sub_items[name];
}

void ParameterizedItem::registerProperty(const QString &name, const QVariant &variant,
                                         const PropertyAttribute &attribute)
{
    if (m_registered_properties.contains(name))
        throw GUIHelpers::Error(
            "ParameterizedItem::registerProperty() -> Error. Already existing property " + name);

    if (m_property_attribute.contains(name))
        throw GUIHelpers::Error(
            "ParameterizedItem::registerProperty() -> Error. Already existing attribute " + name);

    m_registered_properties << name;
    m_property_attribute[name] = attribute;

    setProperty(name.toUtf8().constData(), variant);
}

void ParameterizedItem::setRegisteredProperty(const QString &name, const QVariant &variant)
{
    // check if variant of previous property coincides with new one
    QVariant previous_variant = getRegisteredProperty(name);
    if (GUIHelpers::getVariantType(previous_variant) != GUIHelpers::getVariantType(variant)) {
        qDebug() << "ParameterizedItem::setRegisteredProperty() -> Error. Type of previous and new "
                    "variant does not coincide.";
        qDebug() << "New variant" << variant << ", previous " << previous_variant;
        throw GUIHelpers::Error("ParameterizedItem::setRegisteredProperty() -> Error. Type of "
                                "previous and new variant does not coincide.");
    }

    setProperty(name.toUtf8().constData(), variant);
}

QVariant ParameterizedItem::getRegisteredProperty(const QString &name) const
{
    if (!m_registered_properties.contains(name))
        throw GUIHelpers::Error(
            "ParameterizedItem::getRegisteredProperty() -> Error. Unknown property '" + name
            + "', item '" + modelType() + "'");


    // for debugging purpose
//    PropertyAttribute attribute = getPropertyAttribute(name);
//    if(attribute.getAppearance() == PropertyAttribute::DISABLED) {
//        throw GUIHelpers::Error("ParameterizedItem::getRegisteredProperty() -> Logic Error? "
//            "You are trying to get DISABLED property with name '" +name +"', model " + modelType());
//    }

    return property(name.toUtf8().constData());
}

void ParameterizedItem::removeRegisteredProperty(const QString &name)
{
    if(isRegisteredProperty(name)) {
        qDebug() << "ParameterizedItem::removeRegisteredProperty()" << name;
        if (m_sub_items.contains(name)) {
            qDebug()
                << "       ParameterizedItem::removeRegisteredProperty() -> Removing property iteme";
            delete m_sub_items[name];
            m_sub_items.remove(name);
        }
        m_registered_properties.removeOne(name);
        m_property_attribute.remove(name);
        setProperty(name.toUtf8().constData(), QVariant());
    }
}

PropertyAttribute ParameterizedItem::getPropertyAttribute(const QString &name) const
{
    if (!m_registered_properties.contains(name))
        throw GUIHelpers::Error(
            "ParameterizedItem::getPropertyAttribute() -> Error. Unknown property " + name + " "
            + modelType());

    if (!m_property_attribute.contains(name))
        throw GUIHelpers::Error(
            "ParameterizedItem::getPropertyAttribute() -> Error. Unknown property attribute "
            + name);

    return m_property_attribute[name];
}

void ParameterizedItem::setPropertyAttribute(const QString &name,
                                             const PropertyAttribute &attribute)
{
    if (!m_registered_properties.contains(name))
        throw GUIHelpers::Error(
            "ParameterizedItem::setPropertyAttribute() -> Error. Unknown property " + name);

    if (!m_property_attribute.contains(name))
        throw GUIHelpers::Error(
            "ParameterizedItem::setPropertyAttribute() -> Error. Unknown property attribute "
            + name);

    m_property_attribute[name] = attribute;
}

void ParameterizedItem::setPropertyAppearance(const QString &name,
                                              const PropertyAttribute::EAppearance &appearance)
{
    if (!m_registered_properties.contains(name))
        throw GUIHelpers::Error(
            "ParameterizedItem::setPropertyAppearance() -> Error. Unknown property " + name);

    if (!m_property_attribute.contains(name))
        throw GUIHelpers::Error(
            "ParameterizedItem::setPropertyAppearance() -> Error. Unknown property attribute "
            + name);

    m_property_attribute[name].setAppearance(appearance);
}

void ParameterizedItem::onPropertyChange(const QString &name)
{
    //    qDebug() << "ParameterizedItem::onPropertyChange()" << modelType() << name;
    if (mp_parent)
        mp_parent->onChildPropertyChange();
    emit propertyChanged(name);
}

void ParameterizedItem::onChildPropertyChange()
{
    qDebug() << "ParameterizedItem::onChildPropertyChange()";
    if (mp_parent)
        mp_parent->onChildPropertyChange();
}

void ParameterizedItem::print() const
{
    qDebug() << "--- ParameterizedItem::print() ------------------------------------";
    qDebug() << modelType() << itemName();
    qDebug() << "--- SubItems ---";
    for (QMap<QString, ParameterizedItem *>::const_iterator it = m_sub_items.begin();
         it != m_sub_items.end(); ++it) {
        qDebug() << "   key:" << it.key() << " value->modelType:" << it.value()->modelType();
    }
    qDebug() << "--- Properties ---";
    QList<QByteArray> property_names = dynamicPropertyNames();
    for (int i = 0; i < property_names.length(); ++i) {
        QString name(property_names[i]);
        qDebug() << name << property(name.toUtf8().constData());
    }
    qDebug() << " ";
}

// returns child which should be removed by the model due to over population of children of given
// type
ParameterizedItem *ParameterizedItem::getCandidateForRemoval(ParameterizedItem *new_comer)
{
    if (!new_comer)
        return 0;

    QMap<int, QVector<ParameterizedItem *> > nport_to_nitems;
    foreach (ParameterizedItem *child, m_children) {
        int nport = child->getRegisteredProperty(P_PORT).toInt();
        nport_to_nitems[nport].push_back(child);
    }

    QMap<int, QVector<ParameterizedItem *> >::iterator it = nport_to_nitems.begin();
    while (it != nport_to_nitems.end()) {
        int nport = it.key();
        if (m_port_info.contains(nport)) {
            if (m_port_info[nport].m_item_max_number != 0
                && it.value().size() > m_port_info[nport].m_item_max_number) {
                foreach (ParameterizedItem *item, it.value()) {
                    if (item != new_comer)
                        return item;
                }
            }
        }
        ++it;
    }

    return 0;
}

void ParameterizedItem::setItemPort(ParameterizedItem::PortInfo::EPorts nport)
{
    setRegisteredProperty(P_PORT, nport);
}

void ParameterizedItem::addToValidChildren(const QString &name, PortInfo::EPorts nport,
                                           int nmax_items)
{
    m_valid_children.append(name);

    if (m_port_info.contains(nport)) {
        m_port_info[nport].m_item_names << name;
        m_port_info[nport].m_item_max_number = nmax_items;
    } else {
        m_port_info[nport] = PortInfo(name, nmax_items);
    }
}

QStringList ParameterizedItem::getParameterTreeList() const
{
    QStringList result;
    // add child parameters:
    if (hasChildItems()) {
        for (QList<ParameterizedItem *>::const_iterator it = m_children.begin();
             it != m_children.end(); ++it) {
            QString child_name = (*it)->modelType();
            QStringList child_list = (*it)->getParameterTreeList();
            for (QStringList::const_iterator par_it = child_list.begin();
                 par_it != child_list.end(); ++par_it) {
                QString new_par_name = child_name + QString("/") + *par_it;
                result << new_par_name;
            }
        }
    }
    // add subitem parameters:
    if (m_sub_items.size() > 0) {
        for (QMap<QString, ParameterizedItem *>::const_iterator it = m_sub_items.begin();
             it != m_sub_items.end(); ++it) {
            GroupProperty_t fgp = getRegisteredProperty(it.key()).value<GroupProperty_t>();
            QString subitem_name;
            if (fgp->type()==GroupProperty::SELECTABLE) {
                subitem_name = it.value()->modelType();
            } else {
                subitem_name = it.key();
            }
            ParameterizedItem *subitem = it.value();
            QStringList subitem_list = subitem->getParameterTreeList();
            for (QStringList::const_iterator par_it = subitem_list.begin();
                 par_it != subitem_list.end(); ++par_it) {
                QString new_par_name = subitem_name + QString("/") + *par_it;
                result << new_par_name;
            }
        }
    }
    // add own parameters:
    result << getParameterList();
    return result;
}

//! called when new SubItem appeared
void ParameterizedItem::onSubItemChanged(const QString &propertyName)
{
    connect(m_sub_items[propertyName], SIGNAL(propertyChanged(QString)), this,
            SLOT(processSubItemPropertyChanged(QString)), Qt::UniqueConnection);
    emit subItemChanged(propertyName);
}

void ParameterizedItem::onSubItemPropertyChanged(const QString &property_group,
                                                 const QString &property_name)
{
    emit subItemPropertyChanged(property_group, property_name);
    if (mp_parent)
        mp_parent->onChildPropertyChange();
}

//! called when SubItem change one of its properties
void ParameterizedItem::processSubItemPropertyChanged(const QString &propertyName)
{
    Q_UNUSED(propertyName);
    ParameterizedItem *propertyItem = qobject_cast<ParameterizedItem *>(sender());
    for (QMap<QString, ParameterizedItem *>::iterator it = m_sub_items.begin();
         it != m_sub_items.end(); ++it) {
        if (it.value() == propertyItem) {
            GroupProperty_t group_property
                = getRegisteredProperty(it.key()).value<GroupProperty_t>();
            group_property->setCurrentLabel(propertyItem->getItemLabel());
            onSubItemPropertyChanged(it.key(), propertyName);
            return;
        }
    }
    throw GUIHelpers::Error("ParameterizedItem::onSubItemPropertyChanged() ->"
                            " Error. No such propertyItem found");
}

QStringList ParameterizedItem::getParameterList() const
{
    QStringList result;
    QList<QByteArray> property_names = dynamicPropertyNames();
    for (int i = 0; i < property_names.length(); ++i) {
        QString prop_name = QString(property_names[i]);
        PropertyAttribute prop_attribute = getPropertyAttribute(prop_name);
        if (prop_attribute.getAppearance() & PropertyAttribute::HIDDEN) {
            continue;
        }
        QVariant variant = property(prop_name.toUtf8().constData());
        int type = GUIHelpers::getVariantType(variant);
        if (type == QVariant::Double) {
            result << prop_name;
        }
    }
    return result;
}
